from io import BytesIO
from unittest import expectedFailure
from unittest.mock import call, patch, seal

from botocore.exceptions import ClientError
from django.test import override_settings
from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus
from arkindex.ponos.models import FINAL_STATES, State
from arkindex.process.models import Process, ProcessMode, Revision, WorkerVersion
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Right, Role, User


@override_settings(PONOS_LOG_TAIL=42)
class TestAPI(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.page_1 = cls.corpus.elements.get(name="Volume 1, page 1r")
        cls.page_2 = cls.corpus.elements.get(name="Volume 1, page 2r")
        cls.rev = Revision.objects.first()
        cls.process = Process.objects.get(mode=ProcessMode.Workers)
        cls.process.run()
        cls.task1, cls.task2, cls.task3 = cls.process.tasks.all()

        # Brand new user and corpus with no preexisting rights
        new_user = User.objects.create(email="another@user.com")
        new_corpus = Corpus.objects.create(name="corpus the second")
        Right.objects.create(user=new_user, content_object=new_corpus, level=50)
        cls.process2 = Process.objects.create(
            mode=ProcessMode.Workers,
            creator=new_user,
            corpus=new_corpus,
        )
        cls.process2.run()
        cls.task4 = cls.process2.tasks.first()

        cls.dla = WorkerVersion.objects.get(worker__slug="dla")
        cls.recognizer = WorkerVersion.objects.get(worker__slug="reco")

    def test_task_details_requires_login(self):
        with self.assertNumQueries(0):
            resp = self.client.get(reverse("api:task-details", args=[self.task1.id]))
            self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    @patch("arkindex.project.aws.s3")
    def test_task_details_own_task(self, s3_mock):
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {
            "Body": BytesIO(b"Failed successfully")
        }
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"

        with self.assertNumQueries(3):
            resp = self.client.get(
                reverse("api:task-details", args=[self.task1.id]),
                HTTP_AUTHORIZATION=f"Ponos {self.task1.token}",
            )
            self.assertEqual(resp.status_code, status.HTTP_200_OK)

        self.assertDictEqual(
            resp.json(),
            {
                "id": str(self.task1.id),
                "run": 0,
                "depth": 0,
                "slug": "initialisation",
                "state": "unscheduled",
                "parents": [],
                "logs": "Failed successfully",
                "full_log": "http://somewhere",
                "extra_files": {},
                "agent": None,
                "gpu": None,
                "shm_size": None,
            },
        )

        self.assertEqual(s3_mock.Object.call_count, 2)
        self.assertEqual(s3_mock.Object().get.call_count, 1)
        self.assertEqual(s3_mock.Object().get.call_args, call(Range="bytes=-42"))
        self.assertEqual(s3_mock.meta.client.generate_presigned_url.call_count, 1)
        self.assertEqual(
            s3_mock.meta.client.generate_presigned_url.call_args,
            call("get_object", Params={"Bucket": "ponos", "Key": "somelog"}),
        )

    def test_task_details_other_task_no_access_forbidden(self):
        # Make corpus private just for this test
        self.corpus.public = False
        self.corpus.save()

        with self.assertNumQueries(2):
            resp = self.client.get(
                reverse("api:task-details", args=[self.task2.id]),
                HTTP_AUTHORIZATION=f"Ponos {self.task4.token}",
            )
            self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    def test_task_details_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            resp = self.client.get(reverse("api:task-details", args=[self.task1.id]))
            self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    @expectedFailure
    def test_task_details_requires_process_guest(self):
        self.process.creator = self.superuser
        self.process.save()
        self.corpus.memberships.filter(user=self.user).delete()
        self.corpus.public = False
        self.corpus.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(5):
            resp = self.client.get(reverse("api:task-details", args=[self.task1.id]))
            self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    @patch("arkindex.project.aws.s3")
    def test_task_details_process_level_corpus(self, s3_mock):
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"

        self.process.creator = self.superuser
        self.process.save()
        self.client.force_login(self.user)

        for role in Role:
            s3_mock.reset_mock()
            s3_mock.reset_mock()
            # Recreate the BytesIO each time, because its contents get consumed each time the API is called
            s3_mock.Object.return_value.get.return_value = {"Body": BytesIO(b"Failed successfully")}

            with self.subTest(role=role):
                self.corpus.memberships.filter(user=self.user).update(level=role.value)

                with self.assertNumQueries(4):
                    resp = self.client.get(reverse("api:task-details", args=[self.task1.id]))
                    self.assertEqual(resp.status_code, status.HTTP_200_OK)

                self.assertDictEqual(
                    resp.json(),
                    {
                        "id": str(self.task1.id),
                        "run": 0,
                        "depth": 0,
                        "slug": "initialisation",
                        "state": "unscheduled",
                        "parents": [],
                        "logs": "Failed successfully",
                        "full_log": "http://somewhere",
                        "extra_files": {},
                        "agent": None,
                        "gpu": None,
                        "shm_size": None,
                    },
                )

                self.assertEqual(s3_mock.Object.call_count, 2)
                self.assertEqual(s3_mock.Object().get.call_count, 1)
                self.assertEqual(s3_mock.Object().get.call_args, call(Range="bytes=-42"))
                self.assertEqual(s3_mock.meta.client.generate_presigned_url.call_count, 1)
                self.assertEqual(
                    s3_mock.meta.client.generate_presigned_url.call_args,
                    call("get_object", Params={"Bucket": "ponos", "Key": "somelog"}),
                )

    @patch("arkindex.project.aws.s3")
    def test_task_details(self, s3_mock):
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {
            "Body": BytesIO(b"Failed successfully")
        }
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"

        self.client.force_login(self.superuser)
        with self.assertNumQueries(4):
            resp = self.client.get(reverse("api:task-details", args=[self.task1.id]))
            self.assertEqual(resp.status_code, status.HTTP_200_OK)

        self.assertDictEqual(
            resp.json(),
            {
                "id": str(self.task1.id),
                "run": 0,
                "depth": 0,
                "slug": "initialisation",
                "state": "unscheduled",
                "parents": [],
                "logs": "Failed successfully",
                "full_log": "http://somewhere",
                "extra_files": {},
                "agent": None,
                "gpu": None,
                "shm_size": None,
            },
        )

        self.assertEqual(s3_mock.Object.call_count, 2)
        self.assertEqual(s3_mock.Object().get.call_count, 1)
        self.assertEqual(s3_mock.Object().get.call_args, call(Range="bytes=-42"))
        self.assertEqual(s3_mock.meta.client.generate_presigned_url.call_count, 1)
        self.assertEqual(
            s3_mock.meta.client.generate_presigned_url.call_args,
            call("get_object", Params={"Bucket": "ponos", "Key": "somelog"}),
        )

    def test_update_task_requires_login(self):
        with self.assertNumQueries(0):
            resp = self.client.put(
                reverse("api:task-update", args=[self.task1.id]),
                data={"state": State.Stopping.value},
            )
            self.assertEqual(resp.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_update_task_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            resp = self.client.put(
                reverse("api:task-update", args=[self.task1.id]),
                data={"state": State.Stopping.value},
            )
            self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    def test_update_task_forbids_task(self):
        with self.assertNumQueries(0):
            resp = self.client.put(
                reverse("api:task-update", args=[self.task1.id]),
                data={"state": State.Stopping.value},
                HTTP_AUTHORIZATION=f"Ponos {self.task1.token}",
            )
            self.assertEqual(resp.status_code, status.HTTP_401_UNAUTHORIZED)

    @expectedFailure
    def test_update_task_requires_process_admin_corpus(self):
        self.process.creator = self.superuser
        self.process.save()
        self.corpus.public = False
        self.corpus.save()
        self.client.force_login(self.user)

        for role in [None, Role.Guest, Role.Contributor]:
            with self.subTest(role=role):
                self.corpus.memberships.filter(user=self.user).delete()
                if role:
                    self.corpus.memberships.create(user=self.user, level=role.value)

                with self.assertNumQueries(5):
                    resp = self.client.put(
                        reverse("api:task-update", args=[self.task1.id]),
                        data={"state": State.Stopping.value},
                    )
                    self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    def test_update_running_task_state_stopping(self):
        self.task1.state = State.Running
        self.task1.save()
        self.client.force_login(self.superuser)

        with self.assertNumQueries(4):
            resp = self.client.put(
                reverse("api:task-update", args=[self.task1.id]),
                data={"state": State.Stopping.value},
            )
            self.assertEqual(resp.status_code, status.HTTP_200_OK)

        self.assertDictEqual(resp.json(), {
            "id": str(self.task1.id),
            "state": State.Stopping.value,
        })
        self.task1.refresh_from_db()
        self.assertEqual(self.task1.state, State.Stopping)

    def test_update_non_running_task_state_stopping(self):
        states = list(State)
        states.remove(State.Running)
        self.client.force_login(self.superuser)

        for state in states:
            with self.subTest(state=state):
                self.task1.state = state
                self.task1.save()

                with self.assertNumQueries(3):
                    resp = self.client.put(
                        reverse("api:task-update", args=[self.task1.id]),
                        data={"state": State.Stopping.value},
                    )
                    self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(
                    resp.json(),
                    {"state": [f"Transition from state {state} to state Stopping is forbidden."]},
                )

                self.task1.refresh_from_db()
                self.assertEqual(self.task1.state, state)

    def test_update_final_task_state_pending(self):
        self.task1.state = State.Completed
        self.task1.save()
        self.client.force_login(self.superuser)

        with self.assertNumQueries(5):
            resp = self.client.put(
                reverse("api:task-update", args=[self.task1.id]),
                data={"state": State.Pending.value},
            )
            self.assertEqual(resp.status_code, status.HTTP_200_OK)

        self.task1.refresh_from_db()
        self.assertEqual(self.task1.state, State.Pending)
        self.assertIsNone(self.task1.agent)
        self.assertIsNone(self.task1.gpu)

    def test_update_non_final_task_state_pending(self):
        states = set(State) - set(FINAL_STATES)
        self.client.force_login(self.superuser)

        for state in states:
            with self.subTest(state=state):
                self.task1.state = state
                self.task1.save()

                with self.assertNumQueries(3):
                    resp = self.client.put(
                        reverse("api:task-update", args=[self.task1.id]),
                        data={"state": State.Pending.value},
                    )
                    self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertDictEqual(
                    resp.json(),
                    {"state": [f"Transition from state {state} to state Pending is forbidden."]}
                )
                self.task1.refresh_from_db()
                self.assertEqual(self.task1.state, state)

    def test_partial_update_task_from_agent_requires_login(self):
        with self.assertNumQueries(0):
            resp = self.client.patch(
                reverse("api:task-details", args=[self.task1.id]),
                data={"state": State.Completed.value},
            )
            self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    def test_partial_update_task_from_agent_forbids_users(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            resp = self.client.patch(
                reverse("api:task-details", args=[self.task1.id]),
                data={"state": State.Completed.value},
            )
            self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    def test_partial_update_task_from_agent_forbids_task(self):
        with self.assertNumQueries(1):
            resp = self.client.patch(
                reverse("api:task-details", args=[self.task1.id]),
                data={"state": State.Completed.value},
                HTTP_AUTHORIZATION=f"Ponos {self.task1.token}",
            )
            self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    def test_partial_update_task_requires_login(self):
        with self.assertNumQueries(0):
            resp = self.client.patch(
                reverse("api:task-update", args=[self.task1.id]),
                data={"state": State.Stopping.value},
            )
            self.assertEqual(resp.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_partial_update_task_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            resp = self.client.patch(
                reverse("api:task-update", args=[self.task1.id]),
                data={"state": State.Stopping.value},
            )
            self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    def test_partial_update_task_forbids_task(self):
        with self.assertNumQueries(0):
            resp = self.client.patch(
                reverse("api:task-update", args=[self.task1.id]),
                data={"state": State.Stopping.value},
                HTTP_AUTHORIZATION=f"Ponos {self.task1.token}",
            )
            self.assertEqual(resp.status_code, status.HTTP_401_UNAUTHORIZED)

    @expectedFailure
    def test_partial_update_task_requires_process_admin_corpus(self):
        self.process.creator = self.superuser
        self.process.save()
        self.corpus.public = False
        self.corpus.save()
        self.client.force_login(self.user)

        for role in [None, Role.Guest, Role.Contributor]:
            with self.subTest(role=role):
                self.corpus.memberships.filter(user=self.user).delete()
                if role:
                    self.corpus.memberships.create(user=self.user, level=role.value)

                with self.assertNumQueries(5):
                    resp = self.client.patch(
                        reverse("api:task-update", args=[self.task1.id]),
                        data={"state": State.Stopping.value},
                    )
                    self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    def test_partial_update_running_task_state_stopping(self):
        self.task1.state = State.Running
        self.task1.save()
        self.client.force_login(self.superuser)

        with self.assertNumQueries(4):
            resp = self.client.patch(
                reverse("api:task-update", args=[self.task1.id]),
                data={"state": State.Stopping.value},
            )
            self.assertEqual(resp.status_code, status.HTTP_200_OK)

        self.assertDictEqual(resp.json(), {
            "id": str(self.task1.id),
            "state": State.Stopping.value,
        })
        self.task1.refresh_from_db()
        self.assertEqual(self.task1.state, State.Stopping)

    def test_partial_update_non_running_task_state_stopping(self):
        states = list(State)
        states.remove(State.Running)
        self.task1.save()
        self.client.force_login(self.superuser)

        for state in states:
            with self.subTest(state=state):
                self.task1.state = state
                self.task1.save()

                with self.assertNumQueries(3):
                    resp = self.client.patch(
                        reverse("api:task-update", args=[self.task1.id]),
                        data={"state": State.Stopping.value},
                    )
                    self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(
                    resp.json(),
                    {"state": [f"Transition from state {state} to state Stopping is forbidden."]}
                )

                self.task1.refresh_from_db()
                self.assertEqual(self.task1.state, state)

    def test_partial_update_final_task_state_pending(self):
        self.task1.state = State.Completed
        self.task1.save()
        self.client.force_login(self.superuser)

        with self.assertNumQueries(5):
            resp = self.client.patch(
                reverse("api:task-update", args=[self.task1.id]),
                data={"state": State.Pending.value},
            )
            self.assertEqual(resp.status_code, status.HTTP_200_OK)

        self.task1.refresh_from_db()
        self.assertEqual(self.task1.state, State.Pending)
        self.assertIsNone(self.task1.agent)

    def test_partial_update_non_final_task_state_pending(self):
        states = set(State) - set(FINAL_STATES)
        self.task1.save()
        self.client.force_login(self.superuser)

        for state in states:
            with self.subTest(state=state):
                self.task1.state = state
                self.task1.save()

                with self.assertNumQueries(3):
                    resp = self.client.patch(
                        reverse("api:task-update", args=[self.task1.id]),
                        data={"state": State.Pending.value},
                    )
                    self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertDictEqual(
                    resp.json(),
                    {"state": [f"Transition from state {state} to state Pending is forbidden."]}
                )
                self.task1.refresh_from_db()
                self.assertEqual(self.task1.state, state)

    @patch("arkindex.project.aws.s3")
    def test_task_logs_unicode_error(self, s3_mock):
        """
        Ensure the TaskLogs.latest property is able to handle sliced off Unicode characters.
        Since we fetch the latest logs from S3 using `Range: bytes=-N`, sometimes the characters
        can have a missing byte and cause a UnicodeDecodeError, such as U+00A0 (non-breaking space)
        or U+00A9 (copyright symbol).
        """
        with self.assertRaises(UnicodeDecodeError):
            b"\xa0Failed successfully".decode("utf-8")

        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {
            "Body": BytesIO(b"\xa0Failed successfully")
        }
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"

        self.client.force_login(self.superuser)
        with self.assertNumQueries(4):
            resp = self.client.get(reverse("api:task-details", args=[self.task1.id]))
            self.assertEqual(resp.status_code, status.HTTP_200_OK)
        data = resp.json()

        self.assertEqual(data["logs"], "�Failed successfully")

    @patch("arkindex.project.aws.s3")
    def test_task_logs_errors(self, s3_mock):
        """
        Ensure the TaskLogs.latest property handles missing log files and returns an empty string,
        or handles an InvalidRange error returned only by Ceph when the log file exists but has zero bytes.
        """
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.side_effect = ClientError({"Error": {"Code": ""}}, "get_object")
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"
        seal(s3_mock)

        self.client.force_login(self.superuser)

        for code in ["NoSuchKey", "InvalidRange"]:
            with self.subTest(code=code), self.assertNumQueries(4):
                s3_mock.Object.return_value.get.side_effect.response["Error"]["Code"] = str(code)

                resp = self.client.get(reverse("api:task-details", args=[self.task1.id]))
                self.assertEqual(resp.status_code, status.HTTP_200_OK)
                self.assertEqual(resp.json()["logs"], "")
